#include <assert.h>
#include <ctype.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#define MEM_SIZE 2048 /* zu multiplizieren mit sizeof(memnode) */
/* Verbund, der einer freien oder
   belegten Speicherflaeche vorangeht */
typedef struct memnode {
   size_t size;
      /* Groesse der Speicherflaeche, die diesem Verbund
         unmittelbar folgt, jedoch ohne Einberechnung
	 des Speicherbedarfs fuer diesen Verbund */
   struct memnode* next;
      /* verweist bei freien Speicherflaechen zum naechsten
	 Ringelement der freien Speicherflaechen;
	 bei belegten Speicherflaechen wird dieser Verweis
	 geaendert zu dem unmittelbar vorangehenden Speicherelement,
	 egal ob dieses frei oder belegt ist
      */
} memnode;

/* Speicherbereich, der von den folgenden Funktionen
   verwaltet wird */
memnode dynmem[MEM_SIZE] = {
   /* bleibt immer im Ring der freien Speicherflaechen */
   {0, &dynmem[1]},
   /* enthaelt zu Beginn den gesamten freien Speicher */
   {sizeof dynmem - 2*sizeof(memnode), dynmem}
};
memnode* node = dynmem; /* durchlaeuft den Ring der freien Speicherflaechen */
memnode* root = dynmem; /* verweist auf das erste Element im Ring */

/* um nicht zu sehr zu fragmentieren, werden alle zu belegenden
   Speicherflaechen auf ein Vielfaches von ALIGN aufgerundet;
   dies sollte eine Zweierpotenz sein, die gross genug ist,
   um die Ausrichtungsanforderungen aller elementaren Datentypen
   zu befriedigen */
const int ALIGN = sizeof(memnode);

/* ermittelt, ob die Speicherflaeche bei ptr belegt ist */
bool is_allocated(memnode* ptr) {
   /* Das Wurzelelement ist immer frei */
   if (ptr == root) return false;
   /* Wenn der Zeiger vorwaerts zeigt, dann ist es frei */
   if (ptr->next > ptr) return false;
   /* Der Zeiger zeigt jetzt rueckwaerts, denn
      ptr->next == ptr ist nur beim Wurzelelement moeglich */
   /* Wenn nicht auf Wurzel verwiesen wird, ist ptr belegt */
   if (ptr->next != root) return true;
   /* Wenn ptr->next auf die Wurzel verweist, gibt es zwei
      Faelle:
        - Es ist belegt und ptr liegt unmittelbar hinter
	  dem Wurzelelement
        - Es ist frei und ptr zeigt auf die freie Speicherflaeche
	  mit der hoechsten Adresse
   */
   /* Wenn root->next jenseits von ptr zeigt, dann ist ptr belegt */
   if (root->next > ptr) return true;
   /* Wenn die Wurzel das einzige freie Element ist,
      dann ist ptr ebenfalls belegt */
   return root->next == root;
}

/* liefert einen Zeiger auf die p folgende Speicherflaeche,
   unabhaengig davon, ob sie belegt oder frei ist;
   zu beachten ist, dass bei der hoechstgelegenen Speicherflaeche
   dynmem + MEM_SIZE zurueckgeliefert wird */
memnode* successor(memnode* p) {
   return (memnode*) ((char*)(p+1) + p->size);
}

/* entspricht malloc(), aber bedient sich nur aus dynmem[] */
void* my_malloc(size_t size) {
   assert(size >= 0);
   if (size == 0) return 0;
   /* runde die gewuenschte Groesse auf
      das naechste Vielfache von ALIGN */
   if (size % ALIGN) {
      size += ALIGN - size % ALIGN;
   }
   /* beginnend mit node->next suchen wir nach
      einem Element aus dem Ring der freien Speicherflaechen,
      das genuegend Kapazitaet fuer size Bytes bietet;
      dabei verweist ptr auf das Element, das wir
      gerade untersuchen, und prev auf das vorherige Element */
   memnode* prev = node; memnode* ptr = prev->next;
   do {
      if (ptr->size >= size) break; /* passendes Element gefunden */
      prev = ptr; ptr = ptr->next;
   } while (ptr != node); /* bis der Ring durchlaufen ist */
   if (ptr->size < size) return 0; /* Speicher ist ausgegangen */
   if (ptr->size < size + 2*sizeof(memnode)) {
      node = ptr->next; /* "circular first fit" */
      /* entferne ptr aus dem Ring der freien Speicherflaechen */
      prev->next = ptr->next;
      /* suche nach der unmittelbar vorangehenden Speicherflaeche;
         zu beachten ist hier, dass zwischen prev und ptr noch
	 einige belegte Speicherflaechen liegen koennen
      */
      for (memnode* p = prev; p < ptr; p = successor(p)) {
	 prev = p;
      }
      ptr->next = prev;
      return (void*) (ptr+1);
   }
   node = ptr; /* "circular first fit" */
   /* die bisherige Speicherflaeche wird zerlegt
      in einen neue passende Speicherflaeche, die jetzt belegt wird
      und die alte verkleinerte Speicherflaeche, die uebrigbleibt
   */
   /* lege das neue Element an */
   memnode* newnode = (memnode*)((char*)ptr + ptr->size - size);
   newnode->size = size; newnode->next = ptr;
   /* korrigiere den Zeiger der folgenden Speicherflaeche,
      falls sie belegt sein sollte */
   memnode* next = successor(ptr);
   if (next < dynmem + MEM_SIZE && next->next == ptr) {
      next->next = newnode;
   }
   /* reduziere die Groesse des alten Elements
      aus dem Ring der freien Speicherflaechen */
   ptr->size -= size + sizeof(memnode);
   return (void*) (newnode+1);
}

/* Lege die beiden freien Speicherflaechen bei prev und ptr
   zusammen, falls dies moeglich sein sollte */
void join_if_possible(memnode* prev, memnode* ptr) {
   memnode* p = (memnode*)((char*) prev + sizeof(memnode) + prev->size);
   /* Wenn die beiden Speicherflaechen nicht direkt
      hintereinander liegen, koennen wir sie nicht vereinigen */
   if (p != ptr) return;
   /* Das Wurzelelement muss in Ruhe gelassen werden,
      damit immer mindestens ein Element im Ring der freien
      Speicherflaechen verbleibt */
   if (prev->size == 0) return;
   /* Vereinigung der beiden freien Speicherflaechen */
   prev->next = ptr->next;
   prev->size += ptr->size + sizeof(memnode);
   /* korrigiere den Zeiger der folgenden Speicherflaeche,
      falls sie belegt sein sollte */
   memnode* next = successor(ptr);
   if (next < dynmem + MEM_SIZE && next->next == ptr) {
      next->next = prev;
   }
   /* setze node auf prev, falls sie zuvor auf ptr zeigte */
   if (node == ptr) node = prev;
}

/* gibt die Speicherflaeche bei ptr wieder frei;
   ptr muss zuvor von my_malloc() zurueckgeliefert worden sein  */
void my_free(void* ptr) {
   if (!ptr) return;
   memnode* node = (memnode*) ((char*) ptr - sizeof(memnode));
   /* Ueberpruefe, ob node tatsaechlich belegt ist */
   assert(is_allocated(node));
   /* Wir suchen nach dem vorangehenden Element aus
      dem Ring der freien Speicherflaechen */
   memnode* prev = node->next;
   while (is_allocated(prev)) {
      prev = prev->next;
   }
   /* fuege node wieder in den Ring an passender Stelle ein */
   node->next = prev->next; prev->next = node;
   /* Lege zusammenhaengende freie Speicherflaechen wieder
      zusammen, falls dies moeglich ist */
   join_if_possible(node, node->next);
   join_if_possible(prev, node);
}

/* Gibt zur Analyse den gesamten Status von dynmem[] auf
   der Standardausgabe aus;
   hier wird angenommen, dass Zeiger bequem als int
   darstellbar sind --> das ist nicht portabel */
void debug_alloc(void) {
   for (memnode* ptr = root; ptr < dynmem + MEM_SIZE; ptr = successor(ptr)) {
      if (ptr == node) {
	 putchar('*');
      } else {
	 putchar(' ');
      }
      bool allocated = is_allocated(ptr);
      if (allocated) {
	 putchar('A');
      } else {
	 putchar('F');
      }
      putchar(' ');
      printf("%d %5lu --> %d", (int) ptr,
	 (unsigned long) ptr->size, (int)ptr->next);
      if (allocated) {
	 /* Ausgabe der freigebbaren Adresse */
	 printf(" (f %d)", (int)(ptr+1));
      }
      putchar('\n');
   }
}

/* Interaktives Testprogramm */
int main() {
   int intval;
   void* p;

   char cmdbuf[BUFSIZ] = {0};
   while (printf(": "), fgets(cmdbuf, sizeof cmdbuf, stdin)) {
      char* cp = cmdbuf;
      while (*cp && isspace(*cp)) ++cp;
      switch (*cp++) {
	 case 'a':
	    if (sscanf(cp, "%d", &intval) != 1) break;
	    p = my_malloc(intval);
	    if (p) {
	       printf("got %d bytes at %d\n", intval, (int) p);
	    } else {
	       printf("failed\n");
	    }
	    break;
	 case 'f':
	    if (sscanf(cp, "%d", &intval) != 1) break;
	    my_free((void*) intval);
	    break;
	 case 'd':
	    debug_alloc();
	    break;
	 case 'q':
	    return 0;
	 default:
	    printf("supported commands:\n");
	    printf("a n        allocate n bytes\n");
	    printf("d          debugging output\n");
	    printf("f p        free area at p\n");
	    printf("q          quit\n");
	    break;
      }
   }
}
